( Mandelbrot Plots, 15jun07cht )
( A F# Tutorial for SVFIG, June 23, 2007 )

\ Mandelbrot plots allows one to learn:
\       Mandelbrot fractals
\       Integer numerical computation for floating point numbers
\       486 assmbly in F#
\       Plotting color images in Windows
\       Running Floating Point Unit FPU in 486 assembly
\ ASMTEST.F in the F# system has many examples of assembly programming.
\ Please use these examples as templates when you start writing
\ FORTH words in 486 assembler.
\ You probably need also an Intel reference manual to learn what each
\ assembly code does.  I have a copy of Intel486 Processor Family
\ Programmer's Reference Manual.  It is enormously helpful, especially
\ on the FPU machine code.

\ 1. Mandelbrot Plot
\ Mandelbrot plots are based on the iterative equations on a complex number
\ (x,y), starting at (x0,y0):
\         (x[n+1],y[n+1]) = (x0,y0) + (x[n],y[n])**2
\ The real part is:
\         x[n+1] = x0 + x[n]**2 - y[n]**2
\ The imaginary part is:
\         y[n+1] = y0 + 2x[n]*y[n]
\ When (x[n],y[n]) is outside of the circle of radius 2.0, it diverges
\ quickly.
\ Each point (x,y) has a unique n associated with it.  Assign a color
\ to this point and plot all the points in color.

\ 2. Scaled Integer Numbers
\ Numbers are represented in a scaled integer format.  Bit31 is a sign
\ bit and bits30-28 are the integer bits.  The decimal point is between
\ bit28 and bit27.  Therefore, 1.0 is 10000000H and -1.0 is F0000000H..
\ 2.0 is 200000000H and -2.0 is E0000000H.

\ The product of two integers must be scaled to the left by 4 bits so
\ that the decimal point is realigned.  Since the Mandelbrot plot
\ diverges when (x*x+y*y)>4.0, a mask C0000000H can be used to terminate
\ the iterative computation of a point.

\ 3. Constants and Variables

decimal

: break cr .s key 27 = abort" done" ;

768 constant width
512 constant height
create palette 1024 cells allot
create color_table 1024 cells allot

variable s
variable x
variable y
variable dx
variable left
variable right
variable top
variable bottom
variable xx
variable yy
variable iteration
variable max_iteration
127 max_iteration !
variable colors
16 colors !

\ 4.  Forth Programming
\ ii is the iteration computation to repeat the Mandelbrot plot
\ max_iteration times.
\ The most time comsuming computation is the squaring of x and y,
\ and the multiplication of x and y.
\ The pure Forth codetakes 33 seconds to complete one plot of
\ a 768x512 frame.
\ When SQUARE and MULTIPLY are coded in assembly, it takes 17
\ seconds to complete the same plot.
\ Masking x*x+y*y with $C0000000H causes some artifacts in the plot.
\ Most likely, they are due to overflow in additions.
\ 4SHLD, 5SHLD left shift a double integer by 4 or 5 bits.  Only
\    the higher 32 bits are retained.
\ SQUARE, SQUARE_1 scaled squaring of a 32 bit scaled integer
\ MULTIPLY, IPLY_1 scaled multiplication of two scaled integers
\ II iterative computation of a point at (x,y)

code 4shld ( d -- n )
        mov     eax, [ebp]
        add     ebp, 4 #
        shld    eax, ebx, # 4
        ret
        end-code

code 5shld ( d -- n )
        mov     eax, [ebp]
        add     ebp, 4 #
        shld    eax, ebx, # 5
        ret
        end-code

code square_1
        mov     eax, ebx
        mov     edx, ebx
        imul    eax, edx
        shld    eax, edx, # 4
        mov     ebx, edx
        ret
        end-code

code multiply_1
        mov     eax, [ebp]
        add     ebp, 4 #
        mov     edx, ebx
        imul    eax, edx
        shld    eax, edx, # 5
        mov     ebx, edx
        ret
        end-code

: square ( n -- nn )
   abs dup um* 4shld ;
: multiply ( n m -- 2nm )
   2dup xor 0< s !
   swap abs swap abs um* 5shld
   s @ if negate then ;
: ii ( x y -- n )
   2dup y ! x !
   0 iteration !
   max_iteration @
   for aft
      over square_1 dup xx !
      over square_1 dup yy !  \ break
      + $C0000000 and
      if 2drop r> drop iteration @ exit then
      multiply_1 ( 2xy )
      y @ + >r
      xx @ yy @ - x @ + r>
      1 iteration +!
      then
   next
   2drop iteration @
   ;

\ 5. Color Palette
\ PALETTE is an 1024 entry palette table
\ colors are code in 24 bits of a 32 bit word.  Code format is
\ 00000000bbbbbbbbggggggggrrrrrrrrB
\ Each primary color of red, green and blue has an 8 bit value.
\ FILL_PALETTE
\ Palette table is initialized with a rainbow pattern, with the
\ folowing colors at 128th color boundaries:
\ Black, red, orange, yellow, green, cyne, blue. purple, white
\ FILL_COLOR_TABLE
\ Depending on the value in COLORS, that many colors are selected
\ from the 1024 entry Palette table, and fill the COLOR-TABLE.
\ PIX! takes the colors in COLOR_TABLE and plots the Mandelbrat
\ frame.
\ PIX@ and PIX!
\ Calling windows services to read the color code from a pixel and
\ to write a new color code into a pixel.
\ COLORDEMO, PALETTEDEMO, PALETTEDEMO_1
\ Testing various color frames for verifications.
\ GET_COLOR_1 converts a Mandelbrot iterative number n to a color code
\ in the PALETTE_TABLE.

: fill_palette
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap 1+ swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_color_table
   0 color_table
   1023 for
      over cells $FFF and palette + @
      over !
      cell+ swap
      1024 colors @ / + swap
   next
   2drop ;

fill_palette

: colors ( n -- )
   colors !
   fill_color_table
   ;
16 colors

newcanvas

: 3DUP 2 PICK 2 PICK 2 PICK ;

: pix@ ( x y -- color )
  >R >R
  MEMDC R> R> GetPixel
;

: pix! ( x y color -- )
  >R >R >R
  MEMDC R> R> R> SetPixel DROP
;

: pixdemo ( color -- )
  0 SWAP
  160 FOR AFT >R >R 0 R> R>
  240 FOR AFT
        3DUP ( break ) pix!
        >R >R 1+ R> R>
  THEN NEXT  >R >R DROP R> 1+ R>
  THEN NEXT 2DROP
  update
  ;

: colordemo
  0 0 0
  height for aft
     width for aft
        3dup pix!
        >r >r
        dup $2FF <
        if 1+ else drop 0 then
        r> r>
        1+ $FFFFFF and
     then next
     >r dup $1FF <
     if 1+ else drop 0 then
     r>
  then next
  drop 2drop update
  ;

: get_color ( code -- color )
   $3FF and cells palette + @ ;

: palettedemo
  0 0 0
  height for aft
     width for aft
        3dup get_color pix!
        >r >r
        dup $2FF <
        if 1+ else drop 0 then
        r> r>
        1+
     then next
     >r dup $1FF <
     if 1+ else drop 0 then
     r>
  then next
  drop 2drop update
  ;

: get_color_1 ( code -- color )
   $3FF and cells color_table + @ ;

: palettedemo_1
  128 colors ! fill_color_table
  0 0 0
  height for aft
     width for aft
        3dup get_color_1 pix!
        >r >r
        dup $2FF <
        if 1+ else drop 0 then
        r> r>
        1+
     then next
     >r dup $1FF <
     if 1+ else drop 0 then
     r>
  then next
  drop 2drop update
  ;

\ 6.  Tests of Forth Code
\ SS uses the parameters LEFT, RIGHT, TOP and BOTTOM to produce
\ a Mandelbrot plot.
\ However, these parameters are not very easy to use in panning
\ and scaling the Mandelbrot plots.
\ The x and y axis must keep the same scaling factor.
\ The code are kept here for reference purposes.

: init_frame
   $E0000000 left ! $10000000 right !
   $F0000000 top  ! $10000000 bottom ! ;
init_frame

: steps
   right @ left @ - width / dx !
   ;

\ : pix! 2drop drop ;

: ss
  init_frame
  steps
  top @ y !
  0 ( init y )
  height for aft
     left @ x !
     0 ( init x )
     width for aft
        2dup swap
        x @ y @ ii
        get_color_1 pix!
        1+
        dx @ x +!
     then next
     drop update
     1+
     dx @ y +!
  then next
  drop
  ;

\ 7.  Assembly Code

\ OnePoint in assembly replaces ii in FORTH.
\ (AA,BB) are (x0,y0), the point to iterate
\ (x,y) are current coordinates
\ MASK terminates iteration by testing x*x+y*y before scaling
\ SHLD does the double number scaling
\ SSS plots the initial frame using LEFT, RIGHT, TOP, BOTTOM
\ and DX as parameters.
\ Plotting time is reduced to 3 seconds per frame.
\ RR, LL, UU, DD pan the plot right left, up and down.
\ ++ zooms in by a factor of 2.
\ -- zooms out by a factor of 2.
\ SSS is 10x faster than SS.

variable aa
variable bb
hex 10000000 aa !
10000000 bb !

f8000000 constant mask

code OnePoint ( -- n )
        sub     ebp, 4 #
        mov     [ebp], ebx
        push    edi
        push    esi
        mov     ecx, max_iteration
        mov     ebx, ecx
        mov     eax, aa 
        mov     x  eax
        mov     eax, bb 
        mov     y  eax
@@1:
        mov     eax, y 
        mov     edx, eax
        imul    eax, edx                  ( y*y )
        mov     esi, eax
        mov     edi, edx
        mov     eax, x 
        mov     edx, eax
        imul    eax, edx
        push    eax                       ( x*x )
        push    edx
        add     eax, esi
        adc     edx, edi                  ( x*x+y*y )
        and     edx, # mask
        pop     edx
        pop     eax                       ( restore x*x )
        jnz     @@2
        sub     eax, esi                  ( x*x-y*y )
        sbb     edx, edi
        shld    eax, edx, # 4             ( normalize )
        add     edx, aa                   ( a+x*x-y*y )
        mov     eax, x                    ( save x )
        mov     x  edx                    ( store new x )
        mov     edx, y 
        imul    eax, edx                  ( x*y )
        shld    eax, edx, # 5             ( normalize )
        add     edx, bb
        mov     y  edx                    ( b+2xy )
        loop    @@1
@@2:
        sub     ebx, ecx
        pop     esi
        pop     edi
        ret
        end-code


: iii ( x y -- n )
   bb ! aa ! OnePoint ;
   

: init_frame_1
   $D0000000 left ! $30000000 right !
   $E0000000 top  ! $20000000 bottom !
   $200000 dx ! ;
   init_frame_1

: sss
  top @ yy !
  0 ( init y )
  height for aft
     left @ xx !
     0 ( init x )
     width for aft
        2dup swap
        xx @ yy @ iii \ break
        get_color_1 pix!
        1+
        dx @ xx +!
     then next
     drop update
     1+
     dx @ yy +!
  then next
  drop
  ;

  $400000 constant max_dx
$C0000000 constant min_x
$40000000 constant max_x
$D0000000 constant min_y
$30000000 constant max_y

: rr left @ dx @ $30 * -
   dup min_x <
   if drop
   else left !
     right @ dx @ $30 * - right ! sss
   then
   ;

: ll right @ dx @ $30 * +
   max_x over <
   if drop
   else right !
     left @ dx @ $30 * + left ! sss
   then
   ;

: dd top @ dx @ $30 * -
   dup min_y <
   if drop
   else top !
     bottom @ dx @ $30 * - bottom ! sss
   then
   ;

: uu bottom @ dx @ $30 * +
   max_y over <
   if drop
   else bottom !
     top @ dx @ $30 * + top ! sss
   then
   ;

: ++ dx @ 2/ dup
   if dx ! sss
   else drop
   then ;

: -- dx @ 2* max_dx over <
   if drop 
   else dx ! sss
   then ;


\ 8. More Optimizations

\ SSSS plots current frame with current coordinates and colors
\ using CENTER_X, CENTER_Y, and DELTA as parameters.
\ To reduce computation during panning, the current frame is
\ moved by 48 pixels harizontally or veritcally, only the 48
\ pixel stripes on the edges need to be calculated.
\ Plotting time is reduced to 1.5 seconds on panning.
\ RRR, LLL, UUU, DDD
\ Pan frame to the right, left, or up and down
\ +++  expands the plot by a factor of 2
\ --- shrinks the plot by a factor of 2

variable center_x
variable center_y
variable delta
$400000 constant max_delta

: init_frame_2
   $F7000000 center_x !
   0 center_y !
   $100000 delta !
   ;
init_frame_2

: frame_up
  $30 ( screen y )
  height $30 - for aft
     0 ( screen x )
     width for aft
        2dup swap pix@ >r
        2dup swap $30 - r> pix!
        1+
     then next
     drop
     1+
  then next
  drop update
  ;

: frame_down
  height $30 - dup ( screen y )
  for aft
     0 ( screen x )
     width for aft
        2dup swap pix@ >r
        2dup swap $30 + r> pix!
        1+
     then next
     drop
     1-
  then next
  drop update
  ;

: frame_right
  0 ( screen y )
  height for aft
     width $30 - dup ( screen x )
     for aft
        2dup swap pix@ >r
        2dup $30 + swap r> pix!
        1-
     then next
     drop
     1+
  then next
  drop update
  ;

: frame_left
  0 ( screen y )
  height for aft
     $30 ( screen x )
     width $48 - for aft
        2dup swap pix@ >r
        2dup $30 - swap r> pix!
        1+
     then next
     drop
     1+
  then next
  drop update
  ;

: ssss
  center_y @ delta @ height 2 */ - yy !
  center_x @ delta @ width 2 */ - left !
  0 ( screen y )
  height for aft
     left @ xx !
     0 ( screen x )
     width for aft
        2dup swap
        xx @ yy @ iii \ break
        get_color_1 pix!
        1+
        delta @ xx +!
     then next
     drop update
     1+
     delta @ yy +!
  then next
  drop
  ;

: top_stripe
  center_y @ delta @ height 2 */ - yy !
  center_x @ delta @ width 2 */ - left !
  0 ( screen y )
  $30 for
     left @ xx !
     0 ( screen x )
     width for aft
        2dup swap
        xx @ yy @ iii \ break
        get_color_1 pix!
        1+
        delta @ xx +!
     then next
     drop update
     1+
     delta @ yy +!
  next
  drop
  ;

: bottom_stripe
  center_y @ delta @ height $60 - 2 */ + yy !
  center_x @ delta @ width 2 */ - left !
  height $31 - ( screen y )
  $30 for
     left @ xx !
     0 ( screen x )
     width for aft
        2dup swap
        xx @ yy @ iii \ break
        get_color_1 pix!
        1+
        delta @ xx +!
     then next
     drop update
     1+
     delta @ yy +!
  next
  drop
  ;

: left_stripe
  center_y @ delta @ height 2 */ - yy !
  center_x @ delta @ width 2 */ - left !
  0 ( screen y )
  height for aft
     left @ xx !
     0 ( screen x )
     $30 for
        2dup swap
        xx @ yy @ iii \ break
        get_color_1 pix!
        1+
        delta @ xx +!
     next
     drop update
     1+
     delta @ yy +!
  then next
  drop
  ;

: right_stripe
  center_y @ delta @ height 2 */ - yy !
  center_x @ delta @ width $90 - 2 */ + left !
  0 ( screen y )
  height for aft
     left @ xx !
     width $49 - ( screen x )
     $30 for
        2dup swap
        xx @ yy @ iii \ break
        get_color_1 pix!
        1+
        delta @ xx +!
     next
     drop update
     1+
     delta @ yy +!
  then next
  drop
  ;

: range? ( x y -- f )
   abs dup um* nip swap abs dup um* nip + $F8000000 and ;

: rrr center_x @ delta @ $30 * - dup
   center_y @ range?
   if drop
   else center_x ! ssss
   then
   ;

: lll center_x @ delta @ $30 * + dup
   center_y @ range?
   if drop
   else center_x ! ssss
   then
   ;

: uuu center_y @ delta @ $30 * + dup
   center_x @ range?
   if drop
   else center_y ! ssss
   then
   ;

: ddd center_y @ delta @ $30 * - dup
   center_x @ range?
   if drop
   else center_y ! ssss
   then
   ;

: +++ delta @ 2/ dup
   if delta ! ssss
   else drop
   then ;

: --- delta @ 2* max_delta over <
   if drop
   else delta ! ssss
   then ;

: scene_1 $F7000000 center_x ! 0 center_y ! $200000 delta ! ssss ;
: scene_2 $F7000000 center_x ! 0 center_y ! $100000 delta ! ssss ;
: scene_3 $e5000000 center_x ! 0 center_y ! $1 delta ! ssss ;

: rrr center_x @ delta @ $30 * - dup
   center_y @ range?
   if drop
   else frame_right center_x ! left_stripe
   then
   ;

: lll center_x @ delta @ $30 * + dup
   center_y @ range?
   if drop
   else frame_left center_x ! right_stripe
   then
   ;

: uuu center_y @ delta @ $30 * + dup
   center_x @ range?
   if drop
   else frame_up center_y ! bottom_stripe
   then
   ;

: ddd center_y @ delta @ $30 * - dup
   center_x @ range?
   if drop
   else frame_down center_y ! top_stripe
   then
   ;

\ 9.  Mandelbrot on Floating Point Unit

\ Use the Floating Point Unit in Intel processor to do computation.
\ It allows numbers to be represented in 80 bit floating point
\ format, and will increase the depths of the Mandelbrot plots.
\ FPU multiplications are much faster than integer multiplications.
\ When all the computations are done inside FPU, a 2x increase in
\ speed is expected..
\ There are two issues: one is the computation inside FPU, and the
\ other is the initiation of FPU with the Mandelbrot coordinates.
\ The first experiments are converting the existing Mandelbrot
\ program to make use of FPU.
\ The 486 assembler in F# can assemble all the FPU instructions.
\ FPU is a classic stack-based computing device.  All operations
\ are focused on the top register on the stack.
\ 000 initializes the stack by pushing 5 elements on the stack
\ for Mandelbrot iterations.
\ FTEST pushes another set of parameters.
\ T does one series of Mandelbrot iterations.
\ TT does one Mandelbrot iteration using (x,y) on FORTH stack.
\ T is not quite right yet.  It is my first serious FPU program
\ after a long while.

variable integer
variable four
variable minus1
variable minus3
variable two
variable divisor
$100 divisor !
0 integer !
2 two !
-1 minus1 !
-3 minus3 !
4 four !

code 000 ( f: -- delta x0 y0 x y )
        finit
        fld1
        fidiv    divisor
        fldz
        fldz
        fldz
        fldz
        ret
        end-code

code ftest ( f: -- delta x0 y0 x y )
        finit
        fld1
        fidiv    divisor
        fild     integer
        fld      st
        fadd     st, st(1)
        fist     integer
        ret
        end-code

code t ( -- n, f: -- x y )
        finit
        sub     ebp, 4 #
        mov     [ebp], ebx
        mov     ecx, max_iteration
        mov     ebx, ecx
        fild    dword x
        fild    dword y
@@1:    fld     st(1)
        fmul    st(2)   ( x*x )
        fld     st(1)
        fmul    st(2)   ( y*y )
        fld     st(1)
        fadd    st(1)   ( x*x+y*y)
        ficomp  word four
        fstsw   ax
        sahf
        jnb      @@2     ( x*x+y*y=<4 ? )
        fsubp   st(1), st
        fld     st(2)
        fmul    st(2)
        fadd    st, st
        faddp   st(1), st
        faddp   st(1), st
        loop    @@1
@@2:    sub     ebx, ecx
        ret
        ret
        end-code


: tt ( x y -- n )
        y ! x ! t ;

\ 10.  Mandelbrot in Floating Point

\ In the last section, Mandelbrot coordinates are 32-bit integers,
\ FPU can input 32-bit integers.  It can also input and output
\ integers and floating point numbers in 32-bit, 64-bit and 80-bit
\ extended precision formats.
\ Here I am experimenting with these different formats.
\ When reading and writing numbers from and to memory, use the
\ following words to specify the exact format you want:
\       dword   32-bit
\       double  64-bit
\       long    80-bit
\ Coordinates in x,y are 32-bit integers, scaled by the number in
\   DIVISOR, which represents the floating number 1.0.

\ FLOAT_X, FLOAT_Y, FLOAT_X0, FLOAT_Y0 are used to store floating
\   point numbers representing Mandelbrot coordinates
\ FLOAT_D is the increment on coordinates between neighboring pixels.
\ DUMP_ARRAY is used to store intermediate results for debugging.
\ TXY does Mandelbrot iteration given x,y coordinates on the FORTH
\   stack.  The code which are commented out would store intermediate
\   x,y values in the DUMP_ARRAY to be inspected later.  It is very
\   important in helping me correct many non-trivial mistakes.
\ INIT_COORDINATES, INIT_XY initiate x,y,x0,y0, and are not used.
\ TTT, same as TXY, but with debugging code commented out.
\ TTTT uses initial coordinates in TOP and LEFT to produce a
\   Mandelbrot plot.
\ TTTTT uses initial coordinate in CENTER_X and CENTER_Y tp produce
\   a Mandelbrot plot.  It is more laborious, but easier to pan
\   the Mandelbrot plots, and to zoom in or out.
\ JJ uses the FORTH code II to do Mandelbrot iteration to compare
\   with results from TXY during debugging.

create float_x 10 allot
create float_y 10 allot
create float_x0 10 allot
create float_y0 10 allot
create float_d 10 allot
create dump_array 1000 allot
float_x 10 0 fill
float_y 10 0 fill
float_x0 10 0 fill
float_y0 10 0 fill
float_d 10 0 fill

(       fst long float_x    is illegal )
code init_coordinates
        finit
        fld1
        fidiv   dword divisor
        fstp    long float_d
        fldz
        fstp    long float_x
        fldz
        fstp    long float_y
        fild    dword minus1
        fstp    long float_y0
        fild    dword minus3
        fidiv   dword two
        fstp    long float_x0
        ret
        end-code

code init_xy ( x y -- , scaled x and y with divisor and )
                ( save into float_x and float_y )
        mov     eax, [ebp]
        mov     integer eax
        finit
        fild    dword integer
        fidiv   dword divisor
        fstp    double float_x
        mov     integer ebx
        fild    dword integer
        fidiv   dword divisor
        fstp    double float_y
        mov     ebx, 4 [ebp]
        add     ebp, 8 #

        ret
        end-code

code txy ( x y -- n )
\        push    esi
\        mov     esi, 0 #
        mov     eax, [ebp]              ( get x )
        add     ebp, 4 #
        mov     integer eax
        finit
        fild    dword integer
        fidiv   dword divisor           ( x0 )
        mov     integer ebx
        fild    dword integer
        fidiv   dword divisor           ( x0 y0 )
        mov     ecx, max_iteration
        mov     ebx, ecx
        fld     st(1)                   ( x0 y0 x )
        fld     st(1)                   ( x0 y0 x y )
@@1:    fld     st(1)                   ( x0 y0 x y x )
        fmul    st, st                  ( x0 y0 x y x*x )
        fld     st(1)                   ( x0 y0 x y x*x y )
        fmul    st, st                  ( x0 y0 x y x*x y*y )
        fadd    st, st(1)               ( x0 y0 x y x*x x*x+y*y )
        ficom   dword four              ( x0 y0 x y x*x x*x+y*y )
        fstsw   ax
        sahf
        jnb     @@2                     ( x*x+y*y=<4 ? )
        fsub    st, st(1)               ( x0 y0 x y x*x y*y )
        fsubp   st(1), st               ( x0 y0 x y x*x-y*y )
        fld     st(2)                   ( x0 y0 x y x*x-y*y x )
        fmul    st, st(2)               ( x0 y0 x y x*x-y*y xy )
        fadd    st, st                  ( x0 y0 x y x*x-y*y 2xy )
        fadd    st, st(4)               ( x0 y0 x y x*x-y*y y0+2xy )
        fstp    st(2)                   ( x0 y0 x y0+2xy x*x-y*y  )
        fadd    st, st(4)               ( x0 y0 x y0+2xy x0+x*x-y*y )
        fstp    st(2)                   ( x0 y0 x0+x*x-y*y y0+2xy )
\        fld     st(1)                  ( save x to dump_array )
\        fimul   dword divisor          ( debugging )
\        fistp   dump_array [esi]
\        add     esi, 4 #
\        fld     st                     ( save x to dump_array )
\        fimul   dword divisor          ( debugging )
\        fistp   dump_array [esi]
\        add     esi, 4 #
        loop    @@1
@@2:    sub     ebx, ecx
\        pop     esi
        ret
        end-code


: ttt ( x y -- n )
        txy ;
\        init_xy txy
\        $10000000 divisor !            ( debugging )
\        dump_array 1000 0 fill
\        dump_array
\        over for aft cr
\           dup ?
\           cell+ dup ?
\           cell+
\        then next
\        drop
        ;

: tttt
  $10000000 divisor !
  top @ yy !
  0 ( init y )
  height for aft
     left @ xx !
     0 ( init x )
     width for aft
        2dup swap
        xx @ yy @ ttt \ break
        get_color_1 pix!
        1+
        dx @ xx +!
     then next
     drop update
     1+
     dx @ yy +!
  then next
  drop
  ;

: ttttt
  $10000000 divisor !
  center_y @ delta @ height 2 */ - yy !
  center_x @ delta @ width 2 */ - left !
  0 ( screen y )
  height for aft
     left @ xx !
     0 ( screen x )
     width for aft
        2dup swap
        xx @ yy @ ttt \ break
        get_color_1 pix!
        1+
        delta @ xx +!
     then next
     drop update
     1+
     delta @ yy +!
  then next
  drop
  ;

: jj ( x y -- n, to compare results from ttt during debugging )
   2dup y ! x !
   0 iteration !
   max_iteration @
   for aft
      over square_1 dup xx !
      over square_1 dup yy !  \ break
      + $C0000000 and
      if 2drop r> drop iteration @ exit then
      multiply_1 ( 2xy )
      y @ + >r
      xx @ yy @ - x @ + r>
      1 iteration +!
      then
      cr over . dup .
   next
   2drop iteration @
   ;

\ 11.  Mandelbrot in Extended Precision Floating Point

\ Floating point coordinates in the last section are taked from
\ 32-bit numbers x,y and divided by the integer in DIVISOR.  The
\ precision of coordinates is still limited to 32-bits.
\ Here the coordinates are represented by 64-bit integers, scaled
\ 60 bits to the left to produce coordinates in real numbers.
\ FLOAT_X, FLOAT_Y, FLOAT_X0, FLOAT_Y0 and FLOAT_D are all
\ 64-bit double integers.
\ D@, D!, DLSHIFT, DRSHIFT compliment D+ and DNEGATE for
\ 64-bit integer math operations.
\ FLOAT_D*2 and FLOAT_D/2 scales FLOAT_D for zoom-out and zoom-in.
\ DUM* multiplies a double integer and a single interger to get
\   a double integer product.  It is used to calculate the
\   coordinates at the top-left corner from the center coordinates
\   to initiate a Mandelbrot plot.
\ INCREMENT increments the selected coorindates by value in FLOAT_D.
\ XY does Mandelbrot iteration on (FLOAT_X,FLOAT_Y)
\ XX does a Mandelbrot plot.  Center is ar (FLOAT_X0, FLOAT_Y0) and
\   coordinate increment is in FLOAT_D.
\ XXX does a starting Mandelbrot plot.
\ RRR, LLL, UUU, DDD panning Mandelbrot plot.
\ +++ zooms in on a Mandelbrot plot
\ --- zooms out on a Mandelbrot plot
\ scene_1, scene_2, scene_3, scene_4 various points of interests.
\ MANDELBROT  keyboard interface to explore Mandelbrot plots.
\       key     function
\       = (+)   zoom in
\       -       zoom out
\       u       pan up
\       d       pan down
\       r       pan right
\       l       pan left
\       1       scene_1
\       2       scene_2
\       3       scene_3
\       4       scene_4
\       9       starting scene
\       ESC     quit

( use double integers for delta )
( 1.0 = $1000,0000,0000,0000 )
variable justification
$-3c justification !

code D!
        mov     eax, [ebp]
        mov     ecx, 4 [ebp]
        mov     [ebx], ecx
        mov     4 [ebx], eax
        mov     ebx, 8 [ebp]
        lea     ebp, $0c [ebp]
        ret
        end-code

code D@
        lea     ebp, -4 [ebp]
        mov     eax, [ebx]
        mov     [ebp], eax
        mov     ebx, 4 [ebx]
        ret
        end-code

code DLSHIFT ( d -- 2d )
        mov     eax, [ebp]
        shld    eax, ebx, # 1             ( normalize )
        shl     eax, # 1
        mov     [ebp], eax
        ret
        end-code

code DRSHIFT ( d -- 2d )
        mov     eax, [ebp]
        shrd    ebx, eax, # 1             ( normalize )
        shr     ebx, # 1
        mov     [ebp], eax
        ret
        end-code

: float_d*2
        float_d D@
        dup $10000000 <
        if DLSHIFT float_d D!
        else 2drop
        then
        ;

: float_d/2
        float_d D@ 2dup
        or
        if DRSHIFT float_d D!
        else 2drop
        then
        ;

code xy ( -- n )
        lea     ebp, -4 [ebp]
        mov     [ebp], ebx
        finit
        fild    dword justification
        fild    double float_x
        fscale
        fstp    st(1)
        fild    dword justification
        fild    double float_y
        fscale
        fstp    st(1)                   ( x0 y0 )
        mov     ecx, max_iteration
        mov     ebx, ecx
        fld     st(1)                   ( x0 y0 x )
        fld     st(1)                   ( x0 y0 x y )
@@1:    fld     st(1)                   ( x0 y0 x y x )
        fmul    st, st                  ( x0 y0 x y x*x )
        fld     st(1)                   ( x0 y0 x y x*x y )
        fmul    st, st                  ( x0 y0 x y x*x y*y )
        fld     st(1)                   ( x0 y0 x y x*x y*y x*x )
        fadd    st, st(1)               ( x0 y0 x y x*x y*y x*x+y*y )
        ficomp  dword four              ( x0 y0 x y x*x y*y )
        fstsw   ax
        sahf
        jnb     @@2                     ( x*x+y*y=<4 ? )
        fsubp   st(1), st               ( x0 y0 x y x*x-y*y )
        fld     st(2)                   ( x0 y0 x y x*x-y*y x )
        fmul    st, st(2)               ( x0 y0 x y x*x-y*y xy )
        fadd    st, st                  ( x0 y0 x y x*x-y*y 2xy )
        fadd    st, st(4)               ( x0 y0 x y x*x-y*y y0+2xy )
        fstp    st(2)                   ( x0 y0 x y0+2xy x*x-y*y  )
        fadd    st, st(4)               ( x0 y0 x y0+2xy x0+x*x-y*y )
        fstp    st(2)                   ( x0 y0 x0+x*x-y*y y0+2xy )
\        fld     st(1)                  ( save x to dump_array )
\        fimul   dword divisor          ( debugging )
\        fistp   dump_array [esi]
\        add     esi, 4 #
\        fld     st                     ( save x to dump_array )
\        fimul   dword divisor          ( debugging )
\        fistp   dump_array [esi]
\        add     esi, 4 #
        loop    @@1
@@2:    sub     ebx, ecx
\        pop     esi
        ret
        end-code

: dum* ( d u -- d*u )
        2dup um* drop >r
        nip um* r> + ;

: increment ( a -- )
        dup >r D@
        float_d D@ D+
        r> D! ;

: xx
  float_y0 D@ float_d D@ height 2/ dum* dnegate D+ float_y D!
  0 ( screen y )
  height for aft
     float_x0 D@ float_d D@ width 2/ dum* dnegate D+ float_x D!
     0 ( screen x )
     width for aft
        2dup swap xy \ break
        get_color_1 pix!
        1+
        float_x increment
     then next
     drop update
     1+
     float_y increment
  then next
  drop
  ;

: xxx
   0 $200000 float_d D!
   0 0 float_y0 D!
   0 0 float_x0 D! xx ;

: range? ( x y -- f )
   abs dup um* nip swap abs dup um* nip + $F8000000 and ;

: rrr float_x0 cell+ @ float_d cell+ @ $30 * -
   float_y0 cell+ @ range?
   if
   else float_x0 D@ float_d d@ $30 dum* dnegate D+ float_x0 D! xx
   then
   ;

: lll float_x0 cell+ @ float_d cell+ @ $30 * +
   float_y0 cell+ @ range?
   if
   else float_x0 D@ float_d d@ $30 dum* D+ float_x0 D! xx
   then
   ;

: uuu float_y0 cell+ @ float_d cell+ @ $30 * +
   float_x0 cell+ @ range?
   if
   else float_y0 D@ float_d d@ $30 dum* D+ float_y0 D! xx
   then
   ;

: ddd float_y0 cell+ @ float_d cell+ @ $30 * -
   float_x0 cell+ @ range?
   if
   else float_y0 D@ float_d d@ $30 dum* dnegate D+ float_y0 D! xx
   then
   ;

: ---
        float_d D@
        dup $10000000 <
        if DLSHIFT float_d D! xx
        else 2drop
        then
        ;

: +++
        float_d D@ 2dup
        or
        if DRSHIFT float_d D! xx
        else 2drop
        then
        ;

: scene_1 0 $F7000000 float_x0 d! 0 0 float_y0 d! 0 $200000 float_d d! xx ;
: scene_2 0 $F7000000 float_x0 d! 0 0 float_y0 d! 0 $100000 float_d d! xx ;
: scene_3 0 $e5000000 float_x0 d! 0 0 float_y0 d! 0 1 float_d d! xx ;
: scene_4 0 $e0000000 float_x0 d! 0 0 float_y0 d! 1 0 float_d d! xx ;

: mandelbrot
   xxx
   begin
        key
        dup 75 = if uuu then
        dup 64 = if ddd then
        dup 6c = if lll then
        dup 72 = if rrr then
        dup 3d = if +++ then
        dup 2d = if --- then
        dup 31 = if scene_1 then
        dup 32 = if scene_2 then
        dup 33 = if scene_3 then
        dup 34 = if scene_4 then
        dup 39 = if xxx then
        1b =
   until
   ;

\ 12.  Epilog
\ I was expecting that this Mandelbrot plot could zoom 2**60 times.
\ The tests show that at 2**40, the plots fail to show more details
\ and the fine structures round off.  Perhap you can tell me what
\ goes wrong.

